# Simple LSTM network implemented with PyTorch
# Data: 1s sampled and low pass filtered data

import torch
import torch.nn as nn
torch.set_num_threads(20)
from datetime import datetime
import h5py
import os
import numpy as np
import matplotlib.pyplot as plt
import joblib
from sklearn.preprocessing import RobustScaler, MinMaxScaler
from torch.utils.data import TensorDataset, DataLoader
import torch.optim as optim
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from imv_lstm_networks import IMVFullLSTM
from captum.attr import Saliency
run_start = datetime.now()
# setting the parameters
# LSTM:
n_samp = 10 # number of timesteps per feature (currently needs to devide the length of the timeseries)
n_ch = 7
input_dim = n_ch
output_dim = 1
hidden_dim = 48
batch_size = 2048
layer_dim = 1 # number of layers (lstm)
dropout = 0.0
# IMV_LSTM:
n_ch_input = 7
n_ch_target = 1
n_units = 128
# COMMON
model_name = "lstm"
gpu_num = 1
n_epochs = 2000
learning_rate = 1e-4
weight_decay = 0.0
day_len = 24*60*6 # 10 s sampling rate
num_days_train = 50 # number of days to use for training
num_days_test = 3 # number of days to use for testing, in this case this is = num_val_test
use_fake_signal = False
early_stopping = False
amp_signal = False
# setting what device to use
device = torch.device("cuda:%d" % gpu_num if torch.cuda.is_available() else "cpu")
print("Device:", device)
# file locations
dataloc = '/data/wiay/2263373r/Etna/'
tr_filename = 'simdata_train_nonlinear_12000000_6ch_noisy.hdf5'
ts_filename = 'simdata_test_nonlinear_12000000_6ch_noisy.hdf5'
savedir = '/data/www.astro/2263373r/Etna_LSTM/Etna_LSTM+Captum/'
os.mkdir(savedir + 'run_'+str(run_start))
saveloc = savedir + 'run_' + str(run_start)+'/'

#saving the parameters to a txt file
f = open(saveloc + 'params.txt',"w+")
f.write(f"Model: {model_name} \n")
f.write(f"Learning rate: {learning_rate} \n")
f.write(f"Number of epochs: {n_epochs} \n")
f.write(f"Batch size: {batch_size} \n")
f.write(f"Using fake signal: {use_fake_signal} \n")
f.write(f"Amplifying signal: {amp_signal} \n")
f.write(f"Early stopping: {early_stopping} \n")
if model_name == "lstm":
    f.write(f"Number of layers: {layer_dim} \n")
    f.write(f"Dropout: {dropout} \n")
    f.write(f"Weight decay: {weight_decay} \n")
# -------------------------------------FUNCTIONS--------------------------------------------------
class LSTMModel(nn.Module):
    def __init__(self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob):
        super().__init__()

        # Defining the number of layers and the nodes in each layer
        self.hidden_dim = hidden_dim
        self.layer_dim = layer_dim

        # LSTM layers
        self.lstm = nn.LSTM(
            input_dim, hidden_dim, layer_dim, batch_first=True, dropout=dropout_prob
        )

        # Fully connected layer
        self.fc = nn.Linear(hidden_dim, output_dim)

    def forward(self, x):
        # Initializing hidden state for first input with zeros
        h0 = torch.zeros(self.layer_dim, x.size(0), self.hidden_dim).requires_grad_().to(device)

        # Initializing cell state for first input with zeros
        c0 = torch.zeros(self.layer_dim, x.size(0), self.hidden_dim).requires_grad_().to(device)

        # We need to detach as we are doing truncated backpropagation through time (BPTT)
        # If we don't, we'll backprop all the way to the start even after going through another batch
        # Forward propagation by passing in the input, hidden state, and cell state into the model
        out, (hn, cn) = self.lstm(x, (h0.detach(), c0.detach()))

        # Reshaping the outputs in the shape of (batch_size, seq_length, hidden_size)
        # so that it can fit into the fully connected layer
        out = out[:, -1, :]

        # Convert the final state to our desired output shape (batch_size, output_dim)
        out = self.fc(out).to(device)

        return out

class Optimization:
    def __init__(self, model_name, model, loss_fn, optimizer, labels):
        self.model = model
        self.labels = labels
        self.model_name = model_name
        self.loss_fn = loss_fn
        self.optimizer = optimizer
        self.train_losses = []
        self.val_losses = []

    def train_step(self, x, y):
        # Sets model to train mode
        self.model.train()
        # Makes predictions
        if self.model_name == "lstm":
            yhat = self.model(x)
        if self.model_name == "imv_full_lstm":
            yhat, alpha, beta = self.model(x)
        yhat = torch.squeeze(yhat, axis=1)
        # Computes loss
        loss = self.loss_fn(yhat, y)
        # Computes gradients
        loss.backward()
        # Updates parameters and zeroes gradients
        self.optimizer.step()
        self.optimizer.zero_grad()
        # Returns the loss
        return loss.item()
    def train(self, train_loader, val_loader, test_loader,  batch_size=64, n_epochs=50, n_features=1, saveloc=saveloc,
              input_dim=input_dim):
        epochs_no_improve = 0 
        min_val_loss = 100000  # initially setting the minimum loss to infinity
        n_epochs_stop = 6 # tolerance

        for epoch in range(1, n_epochs + 1):
            epoch_start = datetime.now()
            batch_losses = []
            for x_batch, y_batch in train_loader:
                batch_begin = datetime.now()
                if self.model_name == "lstm":
                    x_batch = x_batch.view([batch_size, -1, n_features]).to(device)
                if self.model_name == "imv_full_lstm":
                    x_batch = x_batch.to(device)
                y_batch = y_batch.to(device)
                loss = self.train_step(x_batch, y_batch)
                batch_losses.append(loss)
                batch_end = datetime.now()
            training_loss = np.mean(batch_losses)
            self.train_losses.append(training_loss)

            with torch.no_grad():
                batch_val_losses = []
                self.model.eval()
                for x_val, y_val in val_loader:
                    if self.model_name == "lstm":
                        x_val = x_val.view([batch_size, -1, n_features]).to(device)
                    if self.model_name == "imv_full_lstm":
                        x_val = x_val.to(device)
                    y_val = y_val.to(device)
                    if self.model_name == "lstm":
                        yhat = self.model(x_val)
                    if self.model_name == "imv_full_lstm":
                        yhat, alpha, beta = self.model(x_val)
                    yhat = torch.squeeze(yhat, axis=1)
                    val_loss = self.loss_fn(y_val, yhat).item()
                    batch_val_losses.append(val_loss)
                validation_loss = np.mean(batch_val_losses)
                self.val_losses.append(validation_loss)
            epoch_end = datetime.now()
            epoch_len = epoch_end-epoch_start
            # printing losses
            if (epoch <= 10) | (epoch % 10 == 0):
                print(
                    f"[{epoch}/{n_epochs}] Training loss: {training_loss:.4f}\t "
                    f"Validation loss: {validation_loss:.4f}\t"
                    f"Time elapsed: {epoch_len}"
                )

            if (epoch <= 10) | (epoch % 10 == 0): # every 10 epochs save the loss plot and model
                self.plot_losses()
                torch.save(self.model.state_dict(), (saveloc + 'model.h5'))
            if (epoch == 1) | (epoch % 50 == 0): # every 50 epochs make a test plot
                if self.model_name == "imv_lstm_full":
                     predictions, values, alphas, betas, attrs = self.evaluate(test_loader, n_features=input_dim)
                     self.plot_feat_imp(alphas, betas, attrs)
                self.plot_saliency(attrs)
                metrics = self.calculate_metrics(predictions, values)
                self.plot_test(predictions, values, metrics)
            # early stopping condition
            if early_stopping:
                if validation_loss < min_val_loss:
                    epochs_no_improve = 0
                    min_val_loss = validation_loss
                else:
                    epochs_no_improve += 1

                if epoch > 50 and epochs_no_improve == n_epochs_stop:
                    print('Early stopping!')
                    break

    def evaluate(self, test_loader, batch_size=1, n_features=1):
        with torch.no_grad():
            predictions = []
            values = []
            alphas = []
            betas = []
            saliency = Saliency(self.model)
            attrs = []
            self.model.eval()
            for x_test, y_test in test_loader:
                if self.model_name == "lstm":
                    x_test = x_test.view([batch_size, -1, n_features]).to(device)
                if self.model_name == "imv_full_lstm":
                    x_test = x_test.to(device)
                y_test = y_test.to(device)
                if self.model_name == "lstm":
                    yhat = self.model(x_test)
                if self.model_name == "imv_full_lstm":
                    yhat, alpha, beta = self.model(x_test)
                yhat = torch.squeeze(yhat, axis=1)
                attr = saliency.attribute(x_test, target=None)
                predictions.append(yhat.cpu().detach().numpy())
                values.append(y_test.cpu().detach().numpy())
                if self.model_name == "imv_full_lstm":
                    alphas.append(alpha.detach().cpu().numpy())
                    betas.append(beta.detach().cpu().numpy())
                attrs.append(attr.detach().cpu().numpy())
            predictions = np.array(predictions).flatten()
            values = np.array(values).flatten()
            attrs = np.concatenate(attrs, axis=0)
            if self.model_name == "imv_full_lstm":
                alphas = np.concatenate(alphas, axis=0)
                betas = np.concatenate(betas, axis=0)
                return predictions, values, alphas, betas, attrs
            else:
                return predictions, values, attrs

    def plot_test(self, predictions, values, metrics):
        p = np.array(predictions)
        v = np.array(values)
        t = np.arange(0, len(p)*10, 10) # 10 seconds sampling rate
        plt.figure(figsize=(10, 5))
        plt.plot(t, p, alpha=0.8, label='prediction')
        plt.plot(t, v, label='truth')
        plt.xlabel('time (s)')
        plt.title('Predictions vs Target')
        plt.text(0, 0.5, f"MAE: {metrics['mae']} \n MSE: {metrics['mse']}") 
        plt.legend()
        plt.savefig(saveloc + 'testpred.png')
        plt.close()

    def plot_saliency(self, attrs):
        attrs = np.expand_dims(attrs, axis=-1)
        attrs = attrs.mean(axis=0)
        attrs = attrs[..., 0]
        attrs = attrs.transpose(1,0)
        fig, ax = plt.subplots(figsize=(20, 20))
        ax.imshow(attrs)
        ax.set_xticks(np.arange(10)) # if the depth changes this needs to change
        ax.set_yticks(np.arange(len(self.labels)))
        ax.set_yticklabels(self.labels)
        for i in range(len(self.labels)):
            for j in range(10): # this as well
                ax.text(j, i, attrs[i, j],
                               ha="center", va="center", color="w")
        ax.set_title("Importance of features and timesteps (Captum)")
        plt.savefig(saveloc + 'saliency_captum.png')
        plt.close()

    def plot_feat_imp(self, alphas, betas):
        alphas = alphas.mean(axis=0)
        betas = betas.mean(axis=0)
        alphas = alphas[..., 0]
        betas = betas[..., 0]
        alphas = alphas.transpose(1, 0)
        fig, ax = plt.subplots(figsize=(20, 20))
        ax.imshow(alphas)
        ax.set_xticks(np.arange(10)) # if the depth changes this needs to change
        ax.set_yticks(np.arange(len(self.labels)))
        ax.set_yticklabels(self.labels)
        for i in range(len(self.labels)):
            for j in range(10): # this as well
                ax.text(j, i, round(alphas[i, j], 3),
                               ha="center", va="center", color="w")
        ax.set_title("Importance of features and timesteps")
        plt.savefig(saveloc + 'var_importance_alphas.png')
        plt.close()
        # plotting betas
        plt.figure(figsize=(10, 10))
        plt.title("Feature importance")
        plt.bar(range(len(self.labels)), betas)
        plt.grid()
        plt.xticks(range(len(self.labels)), self.labels, rotation=90)
        plt.savefig(saveloc + 'var_importance_betas.png')
        plt.close()

    def calculate_metrics(self, predictions, values):
        p = predictions
        v = values
        return {'mae': mean_absolute_error(v, p),
                'mse': mean_squared_error(v, p)}

    def plot_losses(self, saveloc=saveloc):
        plt.figure(figsize=(10, 5))
        plt.plot(self.train_losses, label="Training loss")
        plt.plot(self.val_losses, label="Validation loss")
        plt.legend()
        plt.grid()
        plt.ylabel("MSE")
        plt.yscale("log")
        plt.xlabel("Epoch")
        plt.title("Losses")
        plt.savefig(saveloc + 'loss.png')
        plt.close()

def get_model(model, input_dim, hidden_dim, layer_dim, output_dim, dropout, n_ch_input, n_ch_target):
    models = {
        "lstm": LSTMModel,
        "imv_full_lstm": IMVFullLSTM
    }
    if model == "lstm":
        model_params = {'input_dim': input_dim,
                        'hidden_dim': hidden_dim,
                        'layer_dim': layer_dim,
                        'output_dim': output_dim,
                        'dropout_prob': dropout}
    if model == "imv_full_lstm":
        model_params = {'input_dim': n_ch_input,
                        'output_dim': n_ch_target,
                        'n_units': n_units}
    return models.get(model.lower())(**model_params)

def fake_data(x_data):
    t = np.arange(len(x_data[:, 0]))
    s_1 = np.cos(t/10000)
    s_2 = np.sin(t/20000)
    s_3 = np.cos(t/100000)
    s_4 = np.sin(t/400000)
    y = s_1 + s_2 + s_3 + s_4
    y_scaled = y*np.std(x_data[:, 0]/10)
    new_x = x_data[:, 0] + y_scaled
    # scaling the signal
    scaler = RobustScaler()
    y = scaler.fit_transform(y_scaled.reshape(-1, 1))
    y = np.squeeze(y, axis=1)
    joblib.dump(scaler, "igrav_scaler.z")
    new_x = scaler.fit_transform(new_x.reshape(-1, 1))
    new_x = np.squeeze(new_x, axis=1)
    x_data[:, 0] = new_x
    return y, x_data

def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)

def plot_inputs(x_data,y_data,saveloc=saveloc, run_start=run_start):
    t = np.arange(0, np.shape(x_data)[0] * 10, 10)
    plt.figure(figsize=(7,15))
    plt.subplot(np.shape(x_data)[1] + 1, 1, 1)
    plt.plot(t, y_data)
    for i in range(np.shape(x_data)[1]):
        plt.subplot(np.shape(x_data)[1] + 1, 1, i + 2)
        plt.plot(t, x_data[:, i])
    plt.savefig(saveloc+'input_data.png')
    plt.xlabel('time (s)')
    plt.show()
    plt.close()

#---------------------------------LOADING FILES---------------------------------------
# Training data
n = h5py.File(dataloc + tr_filename, 'r')
# channels to use
labels = np.array(n['labels'])
labels = np.append(labels, 'MEMS')

depth = 10
x_data = np.real(np.array(n['x_data']).T)
y_data = np.real(np.array(n['y_data']))
x_data_arr = np.zeros((np.shape(x_data)[0], depth, np.shape(x_data)[1]))
for j in range(np.shape(x_data)[1]): # for each channel
    for i in range(depth): # for each depth value
        insert = np.full(depth-i-1, x_data[0, j]) # insert size decreases with increasing depth
        # set each item to the first value of x_data for each channel
        # inserting the array in the beginning, and then cutting the array short
        x_data_arr[:, i, j] = np.insert(x_data[:, j], 0, insert, axis=0)[:np.shape(x_data_arr)[0]]

# plotting in the input data
plot_inputs(x_data,y_data)

# reshape into (samples, timestep_per_sample, features), features = channels
# data_X = np.reshape(x_data, (int(np.shape(x_data)[0]/n_samp), n_samp, int(np.shape(x_data)[1])))
# data_Y = np.reshape(y_data, (int(np.shape(y_data)[0]/n_samp), n_samp))
train_X = x_data_arr
train_Y = y_data

# Test and Validation data
m = h5py.File(dataloc+ts_filename, 'r')

depth = 10
x_data = np.real(np.array(m['x_data']).T)
y_data = np.real(np.array(m['y_data']))
x_data_arr = np.zeros((np.shape(x_data)[0], depth, np.shape(x_data)[1]))
for j in range(np.shape(x_data)[1]): # for each channel
    for i in range(depth): # for each depth value
        insert = np.full(depth-i-1, x_data[0, j]) # insert size decreases with increasing depth
        # set each item to the first value of x_data for each channel
        # inserting the array in the beginning, and then cutting the array short
        x_data_arr[:, i, j] = np.insert(x_data[:, j], 0, insert, axis=0)[:np.shape(x_data_arr)[0]]

len_val = int(len(y_data)*0.1)
val_X = x_data_arr[:len_val, :, :]
val_Y = y_data[:len_val]
test_X = x_data_arr[len_val:len_val*2, :, :]
test_Y = y_data[len_val:len_val*2]

print(f"Training data shape: \t x = {np.shape(train_X)} \t  y = {np.shape(train_Y)}")
print(f"Validation data shape: \t x = {np.shape(val_X)} \t  y = {np.shape(val_Y)}")
print(f"Test data shape: \t x = {np.shape(test_X)} \t  y = {np.shape(test_Y)}")

f.write("Training data shape: x = "+str(np.shape(train_X))+" y = "+str(np.shape(train_Y))+"\n")
f.write("Validation data shape: x = "+str(np.shape(val_X))+" y = "+str(np.shape(val_Y))+"\n")
f.write("Test data shape: x = "+str(np.shape(test_X))+" y = "+str(np.shape(test_Y))+"\n")
f.close()

train_features = torch.Tensor(train_X)
train_targets = torch.Tensor(train_Y)
val_features = torch.Tensor(val_X)
val_targets = torch.Tensor(val_Y)
test_features = torch.Tensor(test_X)
test_targets = torch.Tensor(test_Y)

train = TensorDataset(train_features, train_targets)
val = TensorDataset(val_features, val_targets)
test = TensorDataset(test_features, test_targets)

# using pytorch Dataloader to split intp batches
# drop_last=True, of the last bit is not a whole batch then drop it
train_loader = DataLoader(train, pin_memory=True, batch_size=batch_size, shuffle=False, drop_last=True)
val_loader = DataLoader(val, pin_memory=True, batch_size=batch_size, shuffle=False, drop_last=True)
test_loader = DataLoader(test, pin_memory=True, batch_size=batch_size, shuffle=False, drop_last=True)
test_loader_one = DataLoader(test, pin_memory=True, batch_size=1, shuffle=False, drop_last=True)

# ---------------------------------------------------------------------------------------------------------

model = get_model(model_name, input_dim, hidden_dim, layer_dim, output_dim, dropout, n_ch_input, n_ch_target)
loss_fn = nn.MSELoss(reduction="mean")
optimizer = optim.Adam(model.parameters(), lr=learning_rate, weight_decay=weight_decay)
model.to(device)
print(model_name)
print("Trainable parameters: "+str(count_parameters(model)))
f  = open(saveloc+"params.txt", "w+")
f.write(f"Trainable parameters: {count_parameters(model)} \n")
f.write(f"Optimizer: {optimizer} \n")
# training
start_train = datetime.now()
opt = Optimization(model_name=model_name, model=model, loss_fn=loss_fn, optimizer=optimizer, labels=labels)
opt.train(train_loader, val_loader, test_loader, batch_size=batch_size, n_epochs=n_epochs, n_features=input_dim)
opt.plot_losses()
end_train = datetime.now()
print("Training time: "+str(end_train-start_train))
f.write("Training time: "+ str(end_train-start_train)+"\n")
# testing
start_test = datetime.now()
predictions, values, alphas, betas= opt.evaluate(test_loader, batch_size=batch_size, n_features=input_dim)
metrics = opt.calculate_metrics(predictions, values)
opt.plot_test(predictions, values, metrics)
print("Testing metrics:")
print("MAE: ", metrics['mae'], " MSE: ", metrics['mse'])
end_test = datetime.now()
print("Testing time: ",  end_test-start_test)
f.write("Testing metrics:")
f.write(f"MAE: {metrics['mae']}, MSE: {metrics['mse']} \n")
f.write("Testing time: "+str(end_test-start_test)+ "\n")
f.close()
